{===(tab=2)========================================================
TSlimSplitter v2.01

Copyright by TAKEDA, Kazuhiko
             mailto:kztakeda@t3.rim.or.jp)
             http://www.t3.rim.or.jp/~kztakeda/

Created date:       1996/09/16/Mon.
Last modified date: 1999/05/08/Sat.
-------------------------------------------------------------------
***ToDo***
[BUG]
(1) Reduce the redundant redraw(twice) after dragging.
    Maybe because of ReleaseCapture??

[IMPROVEMENT]
(1) In a focus of SlimSplitter, can move by a keyboard.
==================================================================}
unit SlimSplt;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls;

type
  TSeparateType = (sepHorizontal, sepVertical);

  TSlimSplitter = class(TCustomPanel)
  private

      { attributes(property) }
    FSrcPane, FDstPane: TPanel;           { Source/Destination pane }
    FSepType: TSeparateType;              { Horz.=True/Vert.=False }
    FCanDragging: Boolean;                { Cannot move when False
                                            (default=Flase) }
    FSrcMinWidth, FSrcMinHeight: Word;    { Minimum width/height of SrcPane(0..65535) }
    FDstMinWidth, FDstMinHeight: Word;    { Minimum width/height of DstPane(0..65535) }
    FGap: Byte;                           { Splitter's gap(0..255) }
    FSplitX, FSplitY: Integer;            { Split size }

      { attributes }
    FDownPt, FCurPt: TPoint;              { Mouse down/current point }
    FForm: TForm;                         { OwnerForm }
    FDragging: Boolean;                   { flag in moving }
    FPostMouseUp: TMouseEvent;            { Event handler called after OnMouseUp }
    FAutoSrcInvalidate: Boolean;          { Flag to invalidate SrcPane when SlimSplitter moves (default=False) }
    FAutoDstInvalidate: Boolean;          { Flag to invalidate DstPane when SlimSplitter moves (default=False) }
    FUpdated: Boolean;                    { flag of OnInitialUpdate }
    FDebugMode: Boolean;                  { in Debug mode, SplitX/Y appears as a Caption }

      { property accessor methods }
    procedure SetSrcPane(src: TPanel);
    procedure SetDstPane(dst: TPanel);
    procedure SetGap(gap: Byte);
    procedure SetSeparateType(sep: TSeparateType);
    procedure SetSplitX(x: Integer);
    procedure SetSplitY(y: Integer);
    function  RestrictSplitX(x: Integer): Integer;
    function  RestrictSplitY(y: Integer): Integer;
    procedure SetCanDragging(dragging: Boolean);

    procedure DrawDragging;               { Draw moving SlimSplitter }
    procedure PrepareDC;                  { Create DeviceContext }
    procedure FreeDC;                     { Delete DeviceContext
                                            Called from EndSizing/CancelSizing }
    procedure ModifyCursor;               { Change cursor }

  protected
    procedure BeginSizing(X, Y: Integer); { Called from MouseDown }
    procedure ChangeSizing(X, Y: Integer);{ Called from MouseMove }
    procedure EndSizing(X, Y: Integer);   { Called from MouseUp }
    procedure CancelSizing;               { Cancel dragging }
    procedure RecalcLayout;               { Re-calculate areas of self, SrcPane and DstPane }

      { override }
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;

      { using to initialize SlimSplitter after Create() }
    procedure SetParent(AParent: TWinControl); override;

      { Catch a notification of deleting SrcPane/DstPane }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

      { Helper function which sends invalidate message to child controls(invalidCtrl) recursively }
    procedure InvalidateRec(invalidCtrl: TControl);

      { To create OnInitialUpdate, override Paint() }
    procedure Paint; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure OnParentResize(Sender: TObject);  { On resizing the Parent, do RecalcLayout() }
    procedure InvalidateSrc;                    { Invalidate a SrcPane and its childs recursively }
    procedure InvalidateDst;                    { Invalidate a DstPane and its childs recursively }
    procedure OnInitialUpdate; virtual;

  published
      { properties defined in the super class }
{   property Align;       }
    property Alignment;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderWidth;
    property BorderStyle;
{   property DragCursor;  }
{   property DragMode;    }
{   property Enabled;     }
    property Caption;
    property Color;
    property Ctl3D;
    property Font;
    property Locked;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
{   property OnClick;     }
{   property OnDblClick;  }
{   property OnDragDrop;  }
{   property OnDragOver;  }
{   property OnEndDrag;   }
{   property OnEnter;     }
{   property OnExit;      }
{   property OnMouseDown; }
{   property OnMouseMove; }
{   property OnMouseUp;   }
{   property OnResize;    }
{   property OnStartDrag; }

      { published properties }
    property SrcPane: TPanel read FSrcPane write SetSrcPane;
    property DstPane: TPanel read FDstPane write SetDstPane;
    property SeparateType: TSeparateType read FSepType write SetSeparateType default sepHorizontal;
    property CanDragging: Boolean read FCanDragging write SetCanDragging default True;
    property DebugMode: Boolean read FDebugMode write FDebugMode default false;
    property SrcMinWidth: Word read FSrcMinWidth write FSrcMinWidth default 0;
    property SrcMinHeight: Word read FSrcMinHeight write FSrcMinHeight default 0;
    property DstMinWidth: Word read FDstMinWidth write FDstMinWidth default 0;
    property DstMinHeight: Word read FDstMinHeight write FDstMinHeight default 0;

    property SplitX: Integer read FSplitX write SetSplitX default 50;
    property SplitY: Integer read FSplitY write SetSplitY default 50;
    property Gap: Byte read FGap write SetGap default 6;
    property PostMouseUp: TMouseEvent read FPostMouseUp write FPostMouseUp;
    property AutoSrcInvalidate: Boolean read FAutoSrcInvalidate write FAutoSrcInvalidate default False;
    property AutoDstInvalidate: Boolean read FAutoDstInvalidate write FAutoDstInvalidate default False;
  end;

procedure Register;

{**************************************************************************}
implementation

procedure Register;
begin
  RegisterComponents('Slim', [TSlimSplitter]);
end;

function CToC(CTo, CFrom: TControl; P: TPoint): TPoint;
begin
  Result := CTo.ScreenToClient(CFrom.ClientToScreen(P));
end;

procedure DoMessage(msg: String);
begin
  Application.MessageBox(PChar(msg), 'TSlimSplitter',
                         MB_OK or MB_APPLMODAL or MB_ICONSTOP);
end;

constructor TSlimSplitter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

    { initialize default properties }
  FUpdated := False;
  FSepType := sepHorizontal;
  Cursor := crHSplit;
  DragCursor := crHSplit;

  FCanDragging := True;
  FSrcMinWidth := 0;
  FSrcMinHeight := 0;
  FDstMinWidth := 0;
  FDstMinHeight := 0;
  FGap := 6;
  FSplitX := 50;
  FSplitY := 50;
  FAutoSrcInvalidate := False;
  FAutoDstInvalidate := False;
  FDebugMode := False;

  Align := alNone;

    { initialize attributes }
  FDragging := False;
  FForm := nil;

    { customize VCL's ControlState }
  ControlStyle := ControlStyle
    + [csCaptureMouse, csFixedWidth, csFixedHeight] { Capture mouse event }
    - [csAcceptsControls, csSetCaption];       { Has no child controls/
                                                 Don't set Caption from Name }
  RecreateWnd;
end;

destructor TSlimSplitter.Destroy;
begin
  if (Parent is TPanel) and Assigned((Parent as TPanel).OnResize) then
    (Parent as TPanel).OnResize := nil;
  inherited Destroy;
end;

procedure TSlimSplitter.SetParent(AParent: TWinControl);
begin
  if AParent = nil then
    inherited SetParent(AParent)
  else if AParent is TPanel then
  begin
    inherited SetParent(AParent);
    FForm := GetParentForm(self) as TForm;
    if FForm = nil then
    begin
      DoMessage('Cannot get a parent form.  SlimSplitter does not work properly.');
      Exit;
    end;
    RecalcLayout;
    (AParent as TPanel).OnResize := OnParentResize;
  end
  else
  begin
    inherited SetParent(AParent);
    DoMessage('TSlimSplitter must be placed on a TPanel.  Delete this SlimSplitter!');
  end;
end;

procedure TSlimSplitter.OnParentResize(Sender: TObject);
begin
  RecalcLayout;
end;

procedure TSlimSplitter.SetSrcPane(src: TPanel);
begin
  if src = nil then
    FSrcPane := nil
  else if (FSrcPane <> src) and (src.Parent = Parent) and (FDstPane <> src) then
  begin
    FSrcPane := src;
    RecalcLayout;
  end;
end;

procedure TSlimSplitter.SetDstPane(dst: TPanel);
begin
  if dst = nil then
    FDstPane := nil
  else if (FDstPane <> dst) and (dst.Parent = Parent) and (FSrcPane <> dst) then
  begin
    FDstPane := dst;
    RecalcLayout;
  end;
end;

procedure TSlimSplitter.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
    if AComponent = FSrcPane then FSrcPane := nil
    else if AComponent = FDstPane then FDstPane := nil;
end;

procedure TSlimSplitter.SetGap(gap: Byte);
begin
  if gap <= 1 then gap := 1;

  if FGap <> gap then
  begin
    FGap := gap;
    if FSepType = sepHorizontal then Width := gap
    else Height := gap;
    RecalcLayout;
  end;
end;

procedure TSlimSplitter.ModifyCursor;
begin
  if FCanDragging then
  begin
    if FSepType = sepHorizontal then
    begin
      Cursor := crHSplit;
      DragCursor := crHSplit;
    end
    else
    begin
      Cursor := crVSplit;
      DragCursor := crVSplit;
    end;
  end
  else
  begin
    Cursor := crDefault;
    DragCursor := crDefault;
  end;
end;

procedure TSlimSplitter.SetSeparateType(sep: TSeparateType);
begin
  if FSepType <> sep then
  begin
    FSepType := sep;
    ModifyCursor;
    RecalcLayout;
  end;
end;

procedure TSlimSplitter.RecalcLayout;
var
  W, H: LongInt;  { Client Height/Width }
  X, Y: LongInt;  { Real FSplitX/Y }
begin
  if Parent = nil then Exit;

  W := Parent.ClientWidth;
  H := Parent.ClientHeight;

  if FSepType = sepHorizontal then
  begin
    X := FSplitX;
    if W < X + FGap then X := W - FGap;

    if FSrcPane <> nil then
      with FSrcPane do
      begin
        Align := alNone;
        SetBounds(0, 0, X, H);
      end;
    Self.SetBounds(X, 0, FGap, H);
    if FDstPane <> nil then
      with FDstPane do
      begin
        Align := alNone;
        SetBounds(X+FGap, 0, W-X-FGap, H);
      end;
  end
  else
  begin
    Y := FSplitY;
    if H < Y + FGap then Y := H - FGap;

    if FSrcPane <> nil then
      with FSrcPane do
      begin
        Align := alNone;
        SetBounds(0, 0, W, Y);
      end;
    Self.SetBounds(0, Y, W, FGap);
    if FDstPane <> nil then
      with FDstPane do
      begin
        Align := alNone;
        SetBounds(0, Y+FGap, W, H-Y-FGap);
      end;
  end;
end;

procedure TSlimSplitter.SetSplitX(x: Integer);
begin
  x := Self.RestrictSplitX(x);
  if FSplitX <> x then
  begin
    FSplitX := x;
    RecalcLayout;
  end;
end;

procedure TSlimSplitter.SetSplitY(y: Integer);
begin
  y := Self.RestrictSplitY(y);
  if FSplitY <> y then
  begin
    FSplitY := y;
    RecalcLayout;
  end;
end;

{ restrict x between a client area }
function TSlimSplitter.RestrictSplitX(x: Integer): Integer;
begin
  Result := x;
  if Result < FSrcMinWidth then Result := FSrcMinWidth;
  if Parent.ClientWidth-FGap- FDstMinWidth < Result then Result := Parent.ClientWidth-FGap-FDstMinWidth;
end;

{ restrict y between a client area }
function TSlimSplitter.RestrictSplitY(y: Integer): Integer;
begin
  Result := y;
  if Result < FSrcMinHeight then Result := FSrcMinHeight;
  if Parent.ClientHeight-FGap-FDstMinHeight < Result then Result := Parent.ClientHeight-FGap-FDstMinHeight;
end;

procedure TSlimSplitter.SetCanDragging(dragging: Boolean);
begin
  if dragging = FCanDragging then Exit;
  FCanDragging := dragging;
  ModifyCursor;
end;

procedure TSlimSplitter.BeginSizing(X, Y: Integer);
begin
  if (FSrcPane = nil) or (FDstPane = nil) then
  begin
    DoMessage('Both SrcPane and DstPane are not set.');
    Exit;
  end;
  if not FCanDragging then Exit;

  FDragging := True;
  SetFocus; { To cancel dragging by an [ESC] key, set focus }

      { drawing }
  PrepareDC;
  FDownPt := Point(X, Y);
  FCurPt := FDownPt;
  DrawDragging;
end;

procedure TSlimSplitter.ChangeSizing(X, Y: Integer);
begin
  if not FDragging then Exit;

  DrawDragging;
  FCurPt := Point(X, Y);
  DrawDragging;
end;

procedure TSlimSplitter.EndSizing(X, Y: Integer);
begin
  if not FDragging then Exit;

  DrawDragging;
  FreeDC;

    { Change SlimSplitter's  position }
  if (SeparateType = sepHorizontal) and (FCurPt.X <> FDownPt.X) then
    SplitX := Self.Left + (FCurPt.X - FDownPt.X)
  else if (FCurPt.Y <> FDownPt.Y) then
    SplitY := Self.Top + (FCurPt.Y - FDownPt.Y);
  FDragging := False;

  if FDebugMode then Caption := '('+IntToStr(SplitX)+','+IntToStr(SplitY)+')';
end;

procedure TSlimSplitter.CancelSizing;
begin
  if FDragging then
  begin
    DrawDragging;
    FreeDC;
    SetCaptureControl(nil);
    FDragging := False;
  end;
end;

{ Create and initialize DeviceContext }
procedure TSlimSplitter.PrepareDC;
var
  TopLeft, BottomRight: TPoint;
begin
  FForm.Canvas.Handle := GetDCEx(FForm.Handle, 0, DCX_CACHE or DCX_CLIPSIBLINGS or DCX_LOCKWINDOWUPDATE);

  TopLeft := Point(0, 0);
  BottomRight := Point(Parent.Width, Parent.Height);
  TopLeft := CToC(FForm, Parent, TopLeft);
  BottomRight := CToC(FForm, Parent, BottomRight);
  IntersectClipRect(FForm.Canvas.Handle, TopLeft.X, TopLeft.Y, BottomRight.X, BottomRight.Y);
  with FForm.Canvas do
  begin
    Pen.Color := Self.Color;
    Pen.Width := FGap;
    Pen.Mode := pmXOR;
  end;
end;

procedure TSlimSplitter.FreeDC;
var
  DC: HDC;
begin
  with FForm do
  begin
    DC := Canvas.Handle;
    Canvas.Handle := 0;
    ReleaseDC(Handle, DC);
  end;
end;

{ Draw dragging splitter }
procedure TSlimSplitter.DrawDragging;
var
  BeginPt, EndPt: TPoint;
  xNew, yNew: Integer;
begin
  if FSepType = sepHorizontal then
  begin
    xNew := Self.Left + FGap div 2 + (FCurPt.X - FDownPt.X);
    if xNew < (FSrcMinWidth + FGap div 2) then xNew := FSrcMinWidth + FGap div 2;
    if (Parent.Width - FGap div 2 - FDstMinWidth) < xNew then xNew := Parent.Width - FGap div 2 - FDstMinWidth;
    BeginPt := Point(xNew, 0);
    EndPt := Point(xNew, Height);
  end
  else
  begin
    yNew := Self.Top + FGap div 2 + (FCurPt.Y - FDownPt.Y);
    if yNew < (FSrcMinHeight + FGap div 2) then yNew := FSrcMinHeight + FGap div 2;
    if (Parent.Height - FGap div 2 - FDstMinHeight) < yNew then yNew := Parent.Height - FGap div 2 - FDstMinHeight;
    BeginPt := Point(0, yNew);
    EndPt := Point(Width, yNew);
  end;
  BeginPt := CToC(FForm, Parent, BeginPt);
  EndPt := CToC(FForm, Parent, EndPt);

  with FForm.Canvas do
  begin
    MoveTo(BeginPt.X, BeginPt.Y);
    LineTo(EndPt.X, EndPt.Y);
  end;
end;

procedure TSlimSplitter.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if Button = mbLeft then BeginSizing(X, Y)
  else if Button = mbRight then CancelSizing;
end;

procedure TSlimSplitter.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if FDragging then ChangeSizing(X, Y);
end;

procedure TSlimSplitter.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if FDragging then EndSizing(X, Y);

  if Assigned(FPostMouseUp) then FPostMouseUp(Self, Button, Shift, X, Y);
end;

{ Cancel draggin by an [ESC] key }
procedure TSlimSplitter.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if FDragging and (Key = VK_ESCAPE) then
    CancelSizing
  else
    inherited KeyDown(Key, Shift);
end;

{ To move/resize SlimSplitter in a design mode, override it }
procedure TSlimSplitter.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  W, H: LongInt;  { Client Height/Width }
begin
    { In a design mode, restrict dragging by SetBounds }
  if csDesigning in ComponentState then
  begin
    if FForm = nil then
    begin
      inherited SetBounds(ALeft, ATop, AWidth, AHeight);
      Exit;
    end;
    if Parent = nil then Exit;

    W := Parent.ClientWidth;
    H := Parent.ClientHeight;

    if FSepType = sepHorizontal then
    begin
      FSplitX := Self.RestrictSplitX(ALeft);
      FGap := AWidth;
      if FSrcPane <> nil then
        FSrcPane.SetBounds(0, 0, FSplitX, H);
      inherited SetBounds(FSplitX, 0, FGap, H);
      if FDstPane <> nil then
        FDstPane.SetBounds(FSplitX+FGap, 0, W-FSplitX-FGap, H);
    end
    else
    begin
      FSplitY := Self.RestrictSplitY(ATop);
      FGap := AHeight;
      if FSrcPane <> nil then
        FSrcPane.SetBounds(0, 0, W, FSplitY);
      inherited SetBounds(0, FSplitY, W, FGap);
      if FDstPane <> nil then
        FDstPane.SetBounds(0, FSplitY+FGap, W, H-FSplitY-FGap);
    end;
  end
  else
    inherited SetBounds(ALeft, ATop, AWidth, AHeight);

  if FAutoSrcInvalidate then InvalidateSrc;
  if FAutoDstInvalidate then InvalidateDst;
end;

procedure TSlimSplitter.InvalidateSrc;
begin
  if FSrcPane <> nil then InvalidateRec(FSrcPane);
end;

procedure TSlimSplitter.InvalidateDst;
begin
  if FDstPane <> nil then InvalidateRec(FDstPane);
end;

{ Send invalidate to an argument TControl and its childs }
procedure TSlimSplitter.InvalidateRec(invalidCtrl: TControl);
var
  i: Integer;
begin
  if invalidCtrl is TWinControl then
    if TWinControl(invalidCtrl).ControlCount <> 0 then
      for i := 0 to TWinControl(invalidCtrl).ControlCount - 1 do
        InvalidateRec(TWinControl(invalidCtrl).Controls[i]);

  invalidCtrl.Invalidate;
end;

procedure TSlimSplitter.Paint;
begin
  if not FUpdated then OnInitialUpdate;
  inherited Paint;
end;

procedure TSlimSplitter.OnInitialUpdate;
begin
  FUpdated := True;
  Self.RecalcLayout;
end;

end.
